<?php

declare(strict_types = 1);

namespace Drupal\meta_entity\Plugin\Validation\Constraint;

use Drupal\Core\DependencyInjection\ContainerInjectionInterface;
use Symfony\Component\DependencyInjection\ContainerAwareInterface;
use Symfony\Component\DependencyInjection\ContainerAwareTrait;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Validator\Constraint;
use Symfony\Component\Validator\ConstraintValidator;

/**
 * Validates the target entity type and bundle.
 *
 * The target entity type and bundle should match the ones that are defined in
 * the mapping for this meta entity type.
 */
class MappedTargetEntityValidator extends ConstraintValidator implements ContainerAwareInterface, ContainerInjectionInterface {

  use ContainerAwareTrait;

  /**
   * Constructs a new validator instance.
   *
   * @param \Symfony\Component\DependencyInjection\ContainerInterface $container
   *   The service container.
   */
  public function __construct(ContainerInterface $container) {
    $this->setContainer($container);
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container): self {
    return new static($container);
  }

  /**
   * {@inheritdoc}
   */
  public function validate($items, Constraint $constraint): void {
    /** @var \Drupal\Core\Entity\ContentEntityInterface $target_entity */
    $target_entity = $items->entity;
    if (empty($target_entity)) {
      return;
    }

    /** @var \Drupal\meta_entity\Plugin\Validation\Constraint\MappedTargetEntity $constraint */
    $repository = $this->container->get($this->container->getParameter('meta_entity.repositories')[$constraint->metaEntityTypeId]);
    $entity_types = $repository->getMetaEntityTypesForBundle($target_entity->getEntityTypeId(), $target_entity->bundle());
    if (!in_array($items->getEntity()->bundle(), $entity_types)) {
      $this->context->buildViolation($constraint->message, [
        '%type' => $target_entity->getEntityType()->getLabel(),
        '%bundle' => $target_entity->bundle(),
        '%meta' => $items->getEntity()->get('type')->entity->label(),
      ])->setCode('MappedTargetEntity')
        ->addViolation();
    }
  }

}
